/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/* Evolution Calendar and Tasks OGo backend
 *
 * Copyright 2004 (C) Alvaro del Castillo <acs@lambdaux.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 * We use the acronyms:
 * - cbp: e_cal_backend_ogo
 * - comp: cal_component/component
 */

#include <config.h>
#include <string.h>
#include <libgnome/gnome-i18n.h>
#include <libedataserver/e-uid.h>
#include <libecal/e-cal-recur.h>
#include <libecal/e-cal-time-util.h>
#include <libecal/e-cal-util.h>
#include "e-cal-backend-ogo.h"
#include <utils/ogo-connection.h>
#include <libedata-cal/e-cal-backend-cache.h>
#include <libedataserver/e-url.h> 

struct _ECalBackendOGoPrivate {
	GMutex *mutex;

	OGoConnection *conn;
	ECalBackendCache *cache;
	
	gchar *uri, *item_path;
	gboolean read_only;
	CalMode mode;
	icaltimezone *default_zone;
	int timeout_id;

	gchar *username;
	gchar *password;
};

static ECalBackendClass *parent_class = NULL;

/* Time interval in milliseconds for obtaining changes from server and refresh the cache. */
#define CACHE_REFRESH_INTERVAL 600000

#define OGO_ETAG_PROPERTY "X-OGO-ETAG"
#define OGO_HREF_PROPERTY "X-OGO-HREF"

static void
e_cal_backend_ogo_notify_error_code (ECalBackendOGo *ogo, OGoConnectionStatus status)
{
	const char *msg;

	/* FIXME: Implement */
#if 0
	g_return_if_fail (E_IS_CAL_BACKEND_GROUPWISE (cbgw));

	msg = e_gw_connection_get_error_message (status);
	if (msg)
#endif
		e_cal_backend_notify_error (E_CAL_BACKEND (ogo), "Something bad happened");

}

/* Is_read_only handler for the file backend */
static ECalBackendSyncStatus
e_cal_backend_ogo_is_read_only (ECalBackendSync *backend, EDataCal *cal, gboolean *read_only)
{
	ECalBackendOGo *ogo;
	
	ogo = E_CAL_BACKEND_OGO(backend);
	*read_only = ogo->priv->read_only;
	
	return GNOME_Evolution_Calendar_Success;
}

/* return email address of the person who opened the calender */
static ECalBackendSyncStatus
e_cal_backend_ogo_get_cal_address (ECalBackendSync *backend, EDataCal *cal, char **address)
{
	g_warning ("%s: implement me!", __PRETTY_FUNCTION__);

	return GNOME_Evolution_Calendar_OtherError;
}

static ECalBackendSyncStatus
e_cal_backend_ogo_get_alarm_email_address (ECalBackendSync *backend, EDataCal *cal, char **address)
{
	/* OGo does not support email based alarms */
	
	*address = NULL;
	
	return GNOME_Evolution_Calendar_Success;
}

static GMutex *mutex = NULL;

/* FIXME: We might leak an icalcomponent here */
static ECalComponent *
create_component (const gchar *href, const gchar *etag,
		  const gchar *contents, icalcomponent_kind kind)
{
	ECalComponent *comp;
	icalcomponent *calendar, *component;
	icalproperty  *prop;

	g_return_val_if_fail (contents != NULL, NULL);
	
	calendar = icalparser_parse_string (contents);
	
	if (!calendar)
		return NULL;
	
	component = icalcomponent_get_first_component (calendar, kind);
	
	prop = icalproperty_new_x (etag);
	icalproperty_set_x_name (prop, OGO_ETAG_PROPERTY);
	icalcomponent_add_property (component, prop);
	
	prop = icalproperty_new_x (href);
	icalproperty_set_x_name (prop, OGO_HREF_PROPERTY);
	icalcomponent_add_property (component, prop);
		
	comp = e_cal_component_new ();

	if (e_cal_component_set_icalcomponent (comp, component)) {
		return comp;
	}

	icalcomponent_free (calendar);
	g_object_unref (comp);

	return NULL;
}

/* Initialy populate the cache from the server */
static OGoConnectionStatus
populate_cache (ECalBackendOGo *ogo)
{
	ECalBackendOGoPrivate *priv;
	OGoConnectionStatus status;
        ECalComponent *comp;
        GList *list = NULL, *l;
	gboolean done = FALSE;
	int cursor = 0;
	icalcomponent_kind kind;
	GList *items;
	
	priv = ogo->priv;
	kind = e_cal_backend_get_kind (E_CAL_BACKEND (ogo));
	
	if (!mutex) {
		mutex = g_mutex_new ();
	}

	g_mutex_lock (mutex);
	
	status = ogo_connection_fetch_items (ogo->priv->conn,
					     ogo->priv->item_path,
					     &items);

	if (status != OGO_CONNECTION_STATUS_OK)	{
		g_message ("could not fetch items, path is: %s\n", ogo->priv->item_path);
		g_mutex_unlock (mutex);
		return status;
	}

	for (list = items; list != NULL; list = list->next) {
		OGoItem *item = list->data;

		comp = create_component (item->uri, item->etag,
					 item->contents,
					 e_cal_backend_get_kind (E_CAL_BACKEND (ogo)));
		if (comp) {
			char *comp_str;

			e_cal_component_commit_sequence (comp);
			if (kind == icalcomponent_isa (e_cal_component_get_icalcomponent (comp))) {
				comp_str = e_cal_component_get_as_string (comp);
				e_cal_backend_notify_object_created (E_CAL_BACKEND (ogo), (const char *) comp_str);
				g_free (comp_str);
			}
			e_cal_backend_cache_put_component (priv->cache, comp);
		}
		g_object_unref (comp);
	}
	 
	ogo_item_list_free (items);
	
	g_mutex_unlock (mutex);
	
	return OGO_CONNECTION_STATUS_OK;
	
}

/* Gets the etag and href from a component */
static void
get_ogo_info (icalcomponent *comp,
	      gchar **href, gchar **etag)
{
	icalproperty *prop;
	int n_found = 0;

	for (prop = icalcomponent_get_first_property (comp, ICAL_X_PROPERTY);
	     prop;
	     prop = icalcomponent_get_next_property (comp, ICAL_X_PROPERTY)) {
		const char *xname;

		xname = icalproperty_get_x_name (prop);
		g_assert (xname != NULL);

		if (strcmp (xname, OGO_HREF_PROPERTY) == 0) {
			if (href)
				*href = g_strdup (icalproperty_get_x (prop));

			n_found ++;
		} else if (strcmp (xname, OGO_ETAG_PROPERTY) == 0) {
			if (etag)
				*etag = g_strdup (icalproperty_get_x (prop));

			n_found ++;
		}
			
		if (n_found == 2)
			return;
	}

	g_assert_not_reached ();
}

/* Creates a hash table so we can lookup etags from the URI quickly */
typedef struct {
	gchar *uri;
	gchar *etag;
	gboolean touched;
} HashItem;

static void
hash_item_free (HashItem *item)
{
	g_free (item->uri);
	g_free (item->etag);
	g_free (item);
}

static GHashTable *
create_item_hash (GList *items)
{
	GHashTable *table;

	table = g_hash_table_new_full (g_str_hash, g_str_equal,
				       NULL, (GDestroyNotify)hash_item_free);

	for (; items != NULL; items = items->next) {
		OGoItem *item = items->data;
		HashItem *hash_item;

		hash_item = g_new (HashItem, 1);
		hash_item->uri = g_strdup (item->uri);
		hash_item->etag = g_strdup (item->etag);
		hash_item->touched = FALSE;
		
		g_hash_table_insert (table, hash_item->uri, hash_item);
	}
	
	return table;
}

static gchar *
get_item_etag_and_href (GHashTable *table, const gchar *uri, gchar **href)
{
	HashItem *item;

	item = g_hash_table_lookup (table, uri);

	if (!item)
		return NULL;

	if (href)
		*href = item->uri;
	
	item->touched = TRUE;

	
	return item->etag;
}

static void
get_new_uris_foreach (gpointer key, gpointer value, gpointer user_data)
{
	GList **list = user_data;
	HashItem *item = value;

	if (!item->touched)
		*list = g_list_prepend (*list, item->uri);
}

static GList *
get_new_uris (GHashTable *table)
{
	GList *new_uris = NULL;

	g_hash_table_foreach (table,
			      get_new_uris_foreach, &new_uris);

	return new_uris;
}

static ECalComponent *
fetch_component (ECalBackendOGo *ogo,
		 const gchar *href,
		 const gchar *etag,
		 icalcomponent_kind kind)
{
	EUri *uri;
	gchar *contents;
	ECalComponent *comp;
	
	/* FIXME: We assume here that the item hrefs point to the same server which might not be true */
	uri = e_uri_new (href);
	ogo_connection_fetch_item_absolute (ogo->priv->conn, uri->path, &contents);
	e_uri_free (uri);

	comp = create_component (href, etag, contents, kind);
	g_free (contents);
	
	return comp;
}

static void
update_deltas (ECalBackendOGo *ogo, const gchar *watch_href, char **watch_href_contents)
{
	OGoConnectionStatus status;
	GList *items, *list;
	GList *cached_items, *new_items;
	GHashTable *item_hash;
	GList *updated_items = NULL;
	ECalBackendOGoPrivate *priv;
	
	priv = ogo->priv;

	status = ogo_connection_list_items (ogo->priv->conn,
					    ogo->priv->item_path,
					    &items);

	if (status != OGO_CONNECTION_STATUS_OK) {
		e_cal_backend_ogo_notify_error_code (ogo, status);
		return;
	}

	item_hash = create_item_hash (items);
	ogo_item_list_free (items);

	cached_items = e_cal_backend_cache_get_components (ogo->priv->cache);
	for (list = cached_items; list != NULL; list = list->next) {
		ECalComponent *comp = E_CAL_COMPONENT (list->data);
		gchar *href, *etag, *cached_etag;
		gchar *real_href, *comp_str;
		const gchar *uid;
		
		get_ogo_info (e_cal_component_get_icalcomponent (comp), &href, &etag);

		cached_etag = get_item_etag_and_href (item_hash, href, &real_href);

		if (!cached_etag) {
			ECalComponentVType vtype;

			vtype = e_cal_component_get_vtype (comp);
			if ((vtype == E_CAL_COMPONENT_EVENT) ||
			    (vtype == E_CAL_COMPONENT_TODO)) {
				e_cal_component_get_uid (comp, &uid);
				
				comp_str = e_cal_component_get_as_string (comp);
				e_cal_backend_notify_object_removed (E_CAL_BACKEND (ogo), 
								     (char *) uid, comp_str, NULL);
				e_cal_backend_cache_remove_component (ogo->priv->cache, (const char *) uid, NULL);
				g_free (comp_str);
			}

		} else if (strcmp (etag, cached_etag) != 0) {
			updated_items = g_list_prepend (updated_items, real_href);
		}

		g_object_unref (comp);
		
		g_free (href);
		g_free (etag);
	}

	g_list_free (cached_items);
	
	/* Fetch updated items */
	for (list = updated_items; list != NULL; list = list->next) {
		ECalComponent *modified_comp, *cache_comp;
		gchar *href = list->data;
		const gchar *uid;
		gchar *cache_comp_str, *modified_comp_str;
		
 		modified_comp = fetch_component (ogo, href,
						 get_item_etag_and_href (item_hash, href, NULL),
						 e_cal_backend_get_kind (E_CAL_BACKEND (ogo)));
		
		if (!modified_comp) {
			g_message ("Coult not fetch modified component");
			continue;
		}

 		e_cal_component_get_uid (modified_comp, &uid);
		
		cache_comp = e_cal_backend_cache_get_component (ogo->priv->cache,
								uid, NULL);
		e_cal_component_commit_sequence (modified_comp);
		e_cal_component_commit_sequence (cache_comp);

		modified_comp_str = e_cal_component_get_as_string (modified_comp);
		e_cal_backend_cache_remove_component (ogo->priv->cache, uid, NULL);
		e_cal_backend_cache_put_component (ogo->priv->cache, modified_comp);
		
		/* If we found our watch_href, we don't do notify that it's changed;
		 * that's done in the function that called this function
		 */
		if (watch_href && strcmp (watch_href, href) == 0) {
			*watch_href_contents = modified_comp_str;
		}
		else {
			cache_comp_str = e_cal_component_get_as_string (cache_comp);
			
			e_cal_backend_notify_object_modified (E_CAL_BACKEND (ogo), cache_comp_str,
							      modified_comp_str);

			g_free (cache_comp_str);
			g_free (modified_comp_str);
		}
		
		g_object_unref (modified_comp);
	}
	g_list_free (updated_items);
	
	/* New items */
	new_items = get_new_uris (item_hash);
	
	for (list = new_items; list != NULL; list = list->next) {
		gchar *href = list->data;
		ECalComponent *new_comp;
		gchar *new_comp_str;
		
 		new_comp = fetch_component (ogo, href,
					    get_item_etag_and_href (item_hash, href, NULL),
					    e_cal_backend_get_kind (E_CAL_BACKEND (ogo)));

		if (!new_comp) {
			g_message ("Coult not fetch new component");
			continue;
		}

		e_cal_component_commit_sequence (new_comp);

		if (!e_cal_backend_cache_put_component (ogo->priv->cache, new_comp))
			g_message ("Could not add the component");

		new_comp_str = e_cal_component_get_as_string (new_comp);	
				
		/* If we found our watch_href, we don't do notify that it's changed;
		 * that's done in the function that called this function
		 */
		if (watch_href && strcmp (watch_href, href) == 0) {
			*watch_href_contents = new_comp_str;
		} else  {
			e_cal_backend_notify_object_created (E_CAL_BACKEND (ogo), new_comp_str);
			g_free (new_comp_str);
		}
	}
	
	g_list_free (new_items);

	g_hash_table_destroy (item_hash);
}

static gboolean
get_deltas (gpointer handle)
{
	update_deltas (E_CAL_BACKEND_OGO (handle), NULL, NULL);
	
	return TRUE;
}

static ECalBackendSyncStatus
cache_init (ECalBackendOGo *ogo)
{
	ECalBackendOGoPrivate *priv = ogo->priv;
	OGoConnectionStatus status;
	icalcomponent_kind kind;
	const char *time_interval_string;
	int time_interval;

	kind = e_cal_backend_get_kind (E_CAL_BACKEND (ogo));

	time_interval = CACHE_REFRESH_INTERVAL;
	time_interval_string = g_getenv ("GETQM_TIME_INTERVAL");
	if (time_interval_string) {
		time_interval = g_ascii_strtod (time_interval_string, NULL);
		time_interval *= (60*1000); 
		
	}

	/* We poke the cache for a default timezone. Its
	 * absence indicates that the cache file has not been
	 * populated before. */
	if (!e_cal_backend_cache_get_marker (priv->cache)) {
		/* Populate the cache for the first time.*/
		/* start a timed polling thread set to 1 minute*/
		status = populate_cache (ogo);
		if (status != OGO_CONNECTION_STATUS_OK) {
			g_warning (G_STRLOC ": Could not populate the cache");
			/*FIXME  why dont we do a notify here */
			return GNOME_Evolution_Calendar_PermissionDenied;
		} else {
			e_cal_backend_cache_set_marker (priv->cache);
			priv->timeout_id = g_timeout_add (time_interval, (GSourceFunc) get_deltas, (gpointer) ogo);
			priv->mode = CAL_MODE_REMOTE;
			return GNOME_Evolution_Calendar_Success;
		}
	} else {
		GList *cache_items = NULL, *l;
		/* notify the ecal about the objects already in cache */
		cache_items = e_cal_backend_cache_get_components (priv->cache);
		
		for (l = cache_items; l; l = g_list_next (l)) {
			ECalComponent *comp = E_CAL_COMPONENT (l->data);
			char *cal_string;

			if (kind == icalcomponent_isa (e_cal_component_get_icalcomponent (comp))) {
				cal_string = e_cal_component_get_as_string (comp);
				e_cal_backend_notify_object_created (E_CAL_BACKEND (ogo), cal_string);
				g_free (cal_string);
			}
			g_object_unref (comp);
		}		
		if (cache_items)
			g_list_free (cache_items);
		
		/* get the deltas from the cache */
 		update_deltas (ogo, NULL, NULL);
		priv->timeout_id = g_timeout_add (time_interval, (GSourceFunc) get_deltas, (gpointer) ogo);
		priv->mode = CAL_MODE_REMOTE;

		return GNOME_Evolution_Calendar_Success;
	}
	

	return GNOME_Evolution_Calendar_Success;
}

static ECalBackendSyncStatus
connect_to_server (ECalBackendOGo *ogo,
		   const gchar    *username,
		   const gchar    *password)
{
	ECalBackendOGoPrivate *priv;
	ESource *source;
	source = e_cal_backend_get_source (E_CAL_BACKEND (ogo));
        char *uri, *cache_uri;
   	const char *port;
	EUri *parsed_uri;
	int i;
	const char *use_ssl;
	GThread *thread;
	GError *error = NULL;

	priv = ogo->priv;
	uri =  e_source_get_uri (source);
	if(uri == NULL)
		return  GNOME_Evolution_Calendar_OtherError;

	parsed_uri = e_uri_new (uri);

	port = e_source_get_property (source, "port");
	if (port == NULL)
		port = "80";
	use_ssl = e_source_get_property (source, "use_ssl");
	if (use_ssl) 
		priv->uri = g_strconcat ("https://", parsed_uri->host,":", port, NULL );
	else 
		priv->uri = g_strconcat ("http://", parsed_uri->host,":", port, NULL );

	e_uri_free (parsed_uri);
	
	priv->conn = ogo_connection_new (priv->uri, username, password);
	priv->item_path = g_strdup (e_source_get_property (source, "item_path"));

	cache_uri = g_strconcat (uri, priv->item_path, NULL);
	g_free (uri);
 	
	priv->cache = e_cal_backend_cache_new (cache_uri);
	g_free (cache_uri);
	
	if (!priv->cache) {
		g_mutex_unlock (priv->mutex);
		e_cal_backend_notify_error (E_CAL_BACKEND (ogo), _("Could not create cache file"));
		return GNOME_Evolution_Calendar_OtherError;
	}
	
	/* spawn a new thread for opening the calendar */
	thread = g_thread_create ((GThreadFunc) cache_init, ogo, FALSE, &error);
	if (!thread) {
		g_warning (G_STRLOC ": %s", error->message);
		g_error_free (error);
		
		e_cal_backend_notify_error (E_CAL_BACKEND (ogo), _("Could not create thread for populating cache"));
		return GNOME_Evolution_Calendar_OtherError;
	}
	
	return GNOME_Evolution_Calendar_Success;
}

/* Open handler for the file backend */
static ECalBackendSyncStatus
e_cal_backend_ogo_open (ECalBackendSync *backend, 
			EDataCal        *cal, 
			gboolean         only_if_exists,
			const char      *username, 
			const char      *password)
{
	ECalBackendOGo *ogo;
	ECalBackendOGoPrivate *priv;
	ECalBackendSyncStatus status;

	ogo = E_CAL_BACKEND_OGO (backend);
	priv = ogo->priv;

	g_mutex_lock (priv->mutex);

	ogo->priv->read_only = FALSE;
	priv->mode = CAL_MODE_LOCAL;

	/* FIXME: no need to set it online here when we implement the online/offline stuff correctly */
	g_message ("Trying to conenct to server: %s, %s\n", username, password);
	status = connect_to_server (ogo, username, password);

	g_mutex_unlock (priv->mutex);

	return status;
}

static ECalBackendSyncStatus
e_cal_backend_ogo_get_ldap_attribute (ECalBackendSync *backend, EDataCal *cal, char **attribute)
{
	/* ldap attribute is specific to Sun ONE connector to get free busy information*/
	/* retun NULL here as group wise backend know how to get free busy information */
	
	*attribute = NULL;
	
	return GNOME_Evolution_Calendar_Success;
}

static ECalBackendSyncStatus
e_cal_backend_ogo_get_static_capabilities (ECalBackendSync *backend, EDataCal *cal, char **capabilities)
{
	*capabilities = g_strdup (CAL_STATIC_CAPABILITY_NO_EMAIL_ALARMS "," \
				  CAL_STATIC_CAPABILITY_ONE_ALARM_ONLY "," \
				  CAL_STATIC_CAPABILITY_REMOVE_ALARMS ","   \
	                          CAL_STATIC_CAPABILITY_NO_THISANDPRIOR "," \
				  CAL_STATIC_CAPABILITY_NO_THISANDFUTURE "," \
				  CAL_STATIC_CAPABILITY_NO_CONV_TO_ASSIGN_TASK "," \
				  CAL_STATIC_CAPABILITY_NO_CONV_TO_RECUR "," \
				  CAL_STATIC_CAPABILITY_SAVE_SCHEDULES);

	return GNOME_Evolution_Calendar_Success;
}

static ECalBackendSyncStatus
e_cal_backend_ogo_remove (ECalBackendSync *backend, EDataCal *cal)
{
	ECalBackendOGo *ogo;
	ECalBackendOGoPrivate *priv;

	ogo = E_CAL_BACKEND_OGO (backend);
	priv = ogo->priv;
	
	/* remove the cache */
	if (priv->cache)
		e_file_cache_remove (E_FILE_CACHE (priv->cache));

	return GNOME_Evolution_Calendar_OtherError;
}

static ECalBackendSyncStatus
e_cal_backend_ogo_create_object (ECalBackendSync *backend, EDataCal *cal, char **calobj, char **uid)
{
	ECalBackendOGo *ogo;
	ECalBackendOGoPrivate *priv;
	icalcomponent *icalcomp, *calendar;
	gchar *uri;
	OGoConnectionStatus status;
	gchar *new_contents;
	ECalComponent *comp;
	const gchar *prev_uid;
	const gchar *tmp;
	
	ogo = E_CAL_BACKEND_OGO (backend);
	priv = ogo->priv;

	/* FIXME: We only support events right now */
	if (e_cal_backend_get_kind (E_CAL_BACKEND (backend)) == ICAL_VTODO_COMPONENT) {
		e_cal_backend_notify_error (E_CAL_BACKEND (backend), _("Currently, creating new tasks is unsupported"));
		return GNOME_Evolution_Calendar_OtherError;
	}
	
	/* check the component for validity */
	icalcomp = icalparser_parse_string (*calobj);
	if (!icalcomp)
		return GNOME_Evolution_Calendar_InvalidObject;

	if (e_cal_backend_get_kind (E_CAL_BACKEND (backend)) != icalcomponent_isa (icalcomp)) {
		icalcomponent_free (icalcomp);
		return GNOME_Evolution_Calendar_InvalidObject;
	}

	calendar = icalcomponent_new (ICAL_VCALENDAR_COMPONENT);
	icalcomponent_add_component (calendar, icalcomponent_new_clone (icalcomp));

	status = ogo_connection_create_item (ogo->priv->conn, ogo->priv->item_path,
					     "new.ics", icalcomponent_as_ical_string (calendar), "text/calendar",
					     &uri);
	icalcomponent_free (calendar);

	if (status != OGO_CONNECTION_STATUS_OK) {
		e_cal_backend_ogo_notify_error_code (ogo, status);
		return GNOME_Evolution_Calendar_OtherError;
	}

	/* This is a hack. Evolution doesn't expect the new object to
	   have a different UID than the one passed in, so we simulate
	   adding the original one and then immediately removing it */
	   
	e_cal_backend_notify_object_created (E_CAL_BACKEND (ogo), *calobj);
	comp = e_cal_component_new_from_string (*calobj);
	e_cal_component_get_uid (comp, &tmp);
	e_cal_backend_notify_object_removed (E_CAL_BACKEND (ogo), tmp, *calobj, NULL);
	g_object_unref (comp);
	
	/* Re-fetch the item */
	update_deltas (ogo, uri, &new_contents);
	comp = e_cal_component_new_from_string (new_contents);
	e_cal_component_get_uid (comp, &tmp);
	*uid = g_strdup (tmp);
	g_object_unref (comp);
	
	*calobj = new_contents;


	return GNOME_Evolution_Calendar_Success;
}

static ECalBackendSyncStatus
e_cal_backend_ogo_modify_object (ECalBackendSync *backend, EDataCal *cal, const char *calobj, 
				 CalObjModType mod, char **old_object, char **new_object)
{
	ECalBackendOGo *ogo;
	ECalBackendOGoPrivate *priv;
	ECalComponent *comp, *cache_comp;
	icalcomponent *icalcomp, *calendar;
	const gchar *uid;
	gchar *href, *etag;
	OGoConnectionStatus status;
	gchar *obj_str;
	gchar *new_contents;

	/* FIXME: We only support events right now */
	if (e_cal_backend_get_kind (E_CAL_BACKEND (backend)) == ICAL_VTODO_COMPONENT) {
		e_cal_backend_notify_error (E_CAL_BACKEND (backend), _("Currently, modifying tasks is unsupported"));
		return GNOME_Evolution_Calendar_OtherError;
	}
	
	*old_object = NULL;
	
	ogo = E_CAL_BACKEND_OGO (backend);
	priv = ogo->priv;
	
	icalcomp = icalparser_parse_string (calobj);
	if (!icalcomp)
		return GNOME_Evolution_Calendar_InvalidObject;

	comp = e_cal_component_new ();
	e_cal_component_set_icalcomponent (comp, icalcomp);

	e_cal_component_get_uid (comp, &uid);
	g_print ("uid is: %s\n", uid);
	cache_comp = e_cal_backend_cache_get_component (priv->cache, uid, NULL);
	if (!cache_comp) {
		g_message ("CRITICAL : Could not find the object in cache");
		return GNOME_Evolution_Calendar_ObjectNotFound;
	}
	
	calendar = icalcomponent_new (ICAL_VCALENDAR_COMPONENT);
	icalcomponent_add_component (calendar, icalcomponent_new_clone (icalcomp));
	
	get_ogo_info (icalcomp, &href, &etag);

	obj_str = icalcomponent_as_ical_string (calendar);

	status = ogo_connection_modify_item (ogo->priv->conn, href, obj_str, etag,
					     "text/calendar");
	
	if (status != OGO_CONNECTION_STATUS_OK) {
		g_object_unref (comp);
		
		return GNOME_Evolution_Calendar_OtherError;
	}
	
	g_free (etag);

	*old_object = e_cal_component_get_as_string (cache_comp);

	e_cal_backend_cache_put_component (priv->cache, comp);
 
	g_object_unref (cache_comp);
	g_object_unref (comp);

	update_deltas (ogo, href, &new_contents);

	*new_object = new_contents;
	g_free (href); 
	
	return GNOME_Evolution_Calendar_Success;
}

static ECalBackendSyncStatus
e_cal_backend_ogo_remove_object (ECalBackendSync *backend, EDataCal *cal,
				 const char *uid, const char *rid,
				 CalObjModType mod, char **old_object, char **object)
{
	ECalComponent *comp;
	ECalBackendOGo *ogo;
	ECalBackendOGoPrivate *priv;
	gchar *href, *etag;
	OGoConnectionStatus status;
	ogo = E_CAL_BACKEND_OGO (backend);
	priv = ogo->priv;

	*old_object = *object = NULL;
	
	comp = e_cal_backend_cache_get_component (ogo->priv->cache,
						  uid, NULL);
	get_ogo_info (e_cal_component_get_icalcomponent (comp), &href, &etag);

	status = ogo_connection_delete_item (ogo->priv->conn, href, etag);
	
	g_free (href);
	g_free (etag);

	if (status != OGO_CONNECTION_STATUS_OK)
		return GNOME_Evolution_Calendar_OtherError;

	*old_object = e_cal_component_get_as_string (comp);
	*object = NULL;
	
	e_cal_backend_cache_remove_component (ogo->priv->cache, uid, NULL);

	return GNOME_Evolution_Calendar_Success;
}

/* Discard_alarm handler for the file backend */
static ECalBackendSyncStatus
e_cal_backend_ogo_discard_alarm (ECalBackendSync *backend, EDataCal *cal, const char *uid, const char *auid)
{
	return GNOME_Evolution_Calendar_OtherError;
}

static icaltimezone *
e_cal_backend_ogo_internal_get_default_timezone (ECalBackend *backend)
{
	/* Ogo server maintains data in UTC  */
	return icaltimezone_get_utc_timezone ();
}

static icaltimezone *
e_cal_backend_ogo_internal_get_timezone (ECalBackend *backend, const char *tzid)
{
	icaltimezone *zone;

	zone = icaltimezone_get_builtin_timezone_from_tzid (tzid);

	if (!zone)
		return icaltimezone_get_utc_timezone();

	return zone;
}

/* Update_objects handler for the file backend. */
static ECalBackendSyncStatus
e_cal_backend_ogo_receive_objects (ECalBackendSync *backend, EDataCal *cal, const char *calobj)
{
	g_warning ("%s: implement me!", __PRETTY_FUNCTION__);

	return GNOME_Evolution_Calendar_OtherError;
}

static ECalBackendSyncStatus
e_cal_backend_ogo_send_objects (ECalBackendSync *backend, EDataCal *cal, const char *calobj, GList **users,
				char **modified_calobj)
{
	g_warning ("%s: implement me!", __PRETTY_FUNCTION__);

	return GNOME_Evolution_Calendar_OtherError;
}

/* Get_object_component handler for the ogo backend */
static ECalBackendSyncStatus
e_cal_backend_ogo_get_object (ECalBackendSync *backend, EDataCal *cal, const char *uid, const char *rid, char **object)
{
	ECalComponent *comp;
	ECalBackendOGoPrivate *priv;
	ECalBackendOGo *ogo = (ECalBackendOGo *) backend;

	priv = ogo->priv;

	g_mutex_lock (priv->mutex);

	/* search the object in the cache */
	comp = e_cal_backend_cache_get_component (priv->cache, uid, rid);
	if (comp) {
		g_mutex_unlock (priv->mutex);

		if (e_cal_backend_get_kind (E_CAL_BACKEND (backend)) ==
		    icalcomponent_isa (e_cal_component_get_icalcomponent (comp)))
			*object = e_cal_component_get_as_string (comp);
		else
			*object = NULL;

		g_object_unref (comp);

		return *object ? GNOME_Evolution_Calendar_Success : GNOME_Evolution_Calendar_ObjectNotFound;
	}

	g_mutex_unlock (priv->mutex);

	/* callers will never have a uuid that is in server but not in cache */
	return GNOME_Evolution_Calendar_ObjectNotFound;
}

/* Get_objects_in_range handler for the ogo backend */
static ECalBackendSyncStatus
e_cal_backend_ogo_get_object_list (ECalBackendSync *backend, EDataCal *cal, const char *sexp, GList **objects)
{
	ECalBackendOGo *ogo;
	ECalBackendOGoPrivate *priv;
        GList *components, *l;
	ECalBackendSExp *cbsexp;
	gboolean search_needed = TRUE;
        
	ogo = E_CAL_BACKEND_OGO (backend);
	priv = ogo->priv;

	g_mutex_lock (priv->mutex);

	g_message (G_STRLOC ": Getting object list (%s)", sexp);

	if (!strcmp (sexp, "#t"))
		search_needed = FALSE;

	cbsexp = e_cal_backend_sexp_new (sexp);
	if (!cbsexp) {
		g_mutex_unlock (priv->mutex);
		return GNOME_Evolution_Calendar_InvalidQuery;
	}

	*objects = NULL;
	components = e_cal_backend_cache_get_components (priv->cache);
        for (l = components; l != NULL; l = l->next) {
                ECalComponent *comp = E_CAL_COMPONENT (l->data);

		if (e_cal_backend_get_kind (E_CAL_BACKEND (backend)) ==
		    icalcomponent_isa (e_cal_component_get_icalcomponent (comp))) {
			if ((!search_needed) ||
			    (e_cal_backend_sexp_match_comp (cbsexp, comp, E_CAL_BACKEND (backend)))) {
				*objects = g_list_append (*objects, e_cal_component_get_as_string (comp));
			}
		}
        }

	g_list_foreach (components, (GFunc) g_object_unref, NULL);
	g_list_free (components);

	g_mutex_unlock (priv->mutex);

	return GNOME_Evolution_Calendar_Success;
}

static ECalBackendSyncStatus
e_cal_backend_ogo_get_default_object (ECalBackendSync *backend, EDataCal *cal, char **object)
{
	
	ECalComponent *comp;
	
        comp = e_cal_component_new ();

	switch (e_cal_backend_get_kind (E_CAL_BACKEND (backend))) {
	case ICAL_VEVENT_COMPONENT:
        	e_cal_component_set_new_vtype (comp, E_CAL_COMPONENT_EVENT);
		break;
	case ICAL_VTODO_COMPONENT:
		e_cal_component_set_new_vtype (comp, E_CAL_COMPONENT_TODO);
		break;
	default:
		g_object_unref (comp);
		return GNOME_Evolution_Calendar_ObjectNotFound;
	}

	*object = e_cal_component_get_as_string (comp);
	g_object_unref (comp);

	return GNOME_Evolution_Calendar_Success;
}

/* Get_timezone_object handler for the ogo backend */
static ECalBackendSyncStatus
e_cal_backend_ogo_get_timezone (ECalBackendSync *backend, EDataCal *cal, const char *tzid, char **object)
{
	ECalBackendOGo *ogo;
        ECalBackendOGoPrivate *priv;
        icaltimezone *zone;
        icalcomponent *icalcomp;

        ogo = E_CAL_BACKEND_OGO (backend);
        priv = ogo->priv;

        g_return_val_if_fail (tzid != NULL, GNOME_Evolution_Calendar_ObjectNotFound);

        if (!strcmp (tzid, "UTC")) {
                zone = icaltimezone_get_utc_timezone ();
        } else {
		zone = icaltimezone_get_builtin_timezone_from_tzid (tzid);
		if (!zone)
			return GNOME_Evolution_Calendar_ObjectNotFound;
        }

        icalcomp = icaltimezone_get_component (zone);
        if (!icalcomp)
                return GNOME_Evolution_Calendar_InvalidObject;
                                                      
        *object = g_strdup (icalcomponent_as_ical_string (icalcomp));

        return GNOME_Evolution_Calendar_Success;
}

/* Add_timezone handler for the ogo backend */
static ECalBackendSyncStatus
e_cal_backend_ogo_add_timezone (ECalBackendSync *backend, EDataCal *cal, const char *tzobj)
{
	icalcomponent *tz_comp;
	ECalBackendOGo *ogo;
	ECalBackendOGoPrivate *priv;

	ogo = (ECalBackendOGo *) backend;

	g_return_val_if_fail (E_IS_CAL_BACKEND_OGO (ogo), GNOME_Evolution_Calendar_OtherError);
	g_return_val_if_fail (tzobj != NULL, GNOME_Evolution_Calendar_OtherError);

	priv = ogo->priv;

	tz_comp = icalparser_parse_string (tzobj);
	if (!tz_comp)
		return GNOME_Evolution_Calendar_InvalidObject;

	if (icalcomponent_isa (tz_comp) == ICAL_VTIMEZONE_COMPONENT) {
		icaltimezone *zone;

		zone = icaltimezone_new ();
		icaltimezone_set_component (zone, tz_comp);
		if (e_cal_backend_cache_put_timezone (priv->cache, zone) == FALSE) {
			icaltimezone_free (zone, 1);
			return GNOME_Evolution_Calendar_OtherError;
		}
		icaltimezone_free (zone, 1);
	}
	return GNOME_Evolution_Calendar_Success;
}

static ECalBackendSyncStatus
e_cal_backend_ogo_set_default_timezone (ECalBackendSync *backend, EDataCal *cal, const char *tzid)
{
	ECalBackendOGo *ogo;
        ECalBackendOGoPrivate *priv;

	ogo = E_CAL_BACKEND_OGO (backend);
        priv = ogo->priv;
	
	/* Set the default timezone to it. */
	priv->default_zone = icaltimezone_get_builtin_timezone_from_tzid (tzid);

	/* FIXME  write it into the cache*/
	e_cal_backend_cache_put_default_timezone (priv->cache, priv->default_zone);

	return GNOME_Evolution_Calendar_Success;
}

static gboolean
free_busy_instance (ECalComponent *comp,
                    time_t        instance_start,
                    time_t        instance_end,
                    gpointer      data)
{
        icalcomponent *vfb = data;
        icalproperty *prop;
        icalparameter *param;
        struct icalperiodtype ipt;
        icaltimezone *utc_zone;

        utc_zone = icaltimezone_get_utc_timezone ();

        ipt.start = icaltime_from_timet_with_zone (instance_start, FALSE, utc_zone);
        ipt.end = icaltime_from_timet_with_zone (instance_end, FALSE, utc_zone);        ipt.duration = icaldurationtype_null_duration ();

        /* add busy information to the vfb component */
        prop = icalproperty_new (ICAL_FREEBUSY_PROPERTY);
        icalproperty_set_freebusy (prop, ipt);

        param = icalparameter_new_fbtype (ICAL_FBTYPE_BUSY);
        icalproperty_add_parameter (prop, param);

        icalcomponent_add_property (vfb, prop);

        return TRUE;
}

/* function to resolve timezones */
static icaltimezone *
resolve_tzid (const char *tzid, gpointer user_data)
{
        icalcomponent *vcalendar_comp = user_data;

        if (!tzid || !tzid[0])
                return NULL;
        else if (!strcmp (tzid, "UTC"))
                return icaltimezone_get_utc_timezone ();

        return icalcomponent_get_timezone (vcalendar_comp, tzid);
}

static icalcomponent *
create_user_free_busy (ECalBackendOGo *cbogo, const char *address, const char *cn, time_t start, time_t end)
{
        ECalBackendOGoPrivate *priv;
        GList *l;
        icalcomponent *vfb;
        icaltimezone *utc_zone;
        ECalBackendSExp *obj_sexp;
        char *query, *iso_start, *iso_end;

        priv = cbogo->priv;

        /* create the (unique) VFREEBUSY object that we'll return */
        vfb = icalcomponent_new_vfreebusy ();
        if (address != NULL) {
                icalproperty *prop;
                icalparameter *param;

                prop = icalproperty_new_organizer (address);
                if (prop != NULL && cn != NULL) {
                        param = icalparameter_new_cn (cn);
                        icalproperty_add_parameter (prop, param);

                }
                if (prop != NULL)
                        icalcomponent_add_property (vfb, prop);
        }
        utc_zone = icaltimezone_get_utc_timezone ();
        icalcomponent_set_dtstart (vfb, icaltime_from_timet_with_zone (start, FALSE, utc_zone));
        icalcomponent_set_dtend (vfb, icaltime_from_timet_with_zone (end, FALSE, utc_zone));

        /* add all objects in the given interval */
        iso_start = isodate_from_time_t (start);
        iso_end = isodate_from_time_t (end);
        query = g_strdup_printf ("occur-in-time-range? (make-time \"%s\") (make-time \"%s\")",
                                 iso_start, iso_end);
        obj_sexp = e_cal_backend_sexp_new (query);
        g_free (query);
        g_free (iso_start);
        g_free (iso_end);

        if (!obj_sexp)
                return vfb;


        for (l = e_cal_backend_cache_get_components (priv->cache); l; l = l->next) {
                ECalComponent *comp = l->data;
                icalcomponent *icalcomp, *vcalendar_comp;
                icalproperty *prop;

                icalcomp = e_cal_component_get_icalcomponent (comp);
                if (!icalcomp)
                        continue;

                /* If the event is TRANSPARENT, skip it. */
                prop = icalcomponent_get_first_property (icalcomp,
                                                         ICAL_TRANSP_PROPERTY);
                if (prop) {
                        icalproperty_transp transp_val = icalproperty_get_transp (prop);
                        if (transp_val == ICAL_TRANSP_TRANSPARENT ||
                            transp_val == ICAL_TRANSP_TRANSPARENTNOCONFLICT)
                                continue;
                }

                if (!e_cal_backend_sexp_match_comp (obj_sexp, l->data, E_CAL_BACKEND (cbogo)))
                        continue;

                vcalendar_comp = icalcomponent_get_parent (icalcomp);
                e_cal_recur_generate_instances (comp, start, end,
                                                free_busy_instance,
                                                vfb,
                                                resolve_tzid,
                                                vcalendar_comp,
                                                priv->default_zone);
        }

        return vfb;
}

/* Get_free_busy handler for the file backend */
static ECalBackendSyncStatus
e_cal_backend_ogo_get_free_busy (ECalBackendSync *backend, EDataCal *cal, GList *users,
				 time_t start, time_t end, GList **freebusy)
{
	ECalBackendOGo *cbogo;
	ECalBackendOGoPrivate *priv;
	gchar *address, *name;
	icalcomponent *vfb;
	char *calobj;
	GList *l;
	
	cbogo = E_CAL_BACKEND_OGO (backend);

	g_return_val_if_fail (start != -1 && end != -1, GNOME_Evolution_Calendar_InvalidRange);
        g_return_val_if_fail (start <= end, GNOME_Evolution_Calendar_InvalidRange);

	if (users == NULL) {
		if (e_cal_backend_mail_account_get_default (&address, &name)) {
			vfb = create_user_free_busy (cbogo, address, name, start, end);
			calobj = icalcomponent_as_ical_string (vfb);
			g_print ("CALOBJ: %s", calobj);
			*freebusy = g_list_append (*freebusy, g_strdup (calobj));
			icalcomponent_free (vfb);
			g_free (address);
			g_free (name);
		}
	} else {
                for (l = users; l != NULL; l = l->next ) {
                        address = l->data;
                        if (e_cal_backend_mail_account_is_valid (address, &name)) {
                                vfb = create_user_free_busy (cbogo, address, name, start, end);
                                calobj = icalcomponent_as_ical_string (vfb);
				g_print ("CALOBJ: %s", calobj);
                                *freebusy = g_list_append (*freebusy, g_strdup (calobj));
                                icalcomponent_free (vfb);
                                g_free (name);
                        }
                }
	}

	return GNOME_Evolution_Calendar_Success;
}

/* Get_changes handler for the ogo backend */
static ECalBackendSyncStatus
e_cal_backend_ogo_get_changes (ECalBackendSync *backend, EDataCal *cal, const char *change_id,
				     GList **adds, GList **modifies, GList **deletes)
{
	g_warning ("%s: implement me!", __PRETTY_FUNCTION__);

	return GNOME_Evolution_Calendar_OtherError;
}

/* get_query handler for the ogo backend */
static void
e_cal_backend_ogo_start_query (ECalBackend *backend, EDataCalView *query)
{
        ECalBackendSyncStatus status;
	ECalBackendOGo *ogo;
	ECalBackendOGoPrivate *priv;
        GList *objects = NULL;

	ogo = E_CAL_BACKEND_OGO (backend);
	priv = ogo->priv;

        status = e_cal_backend_ogo_get_object_list (E_CAL_BACKEND_SYNC (backend), NULL,
						    e_data_cal_view_get_text (query), &objects);
        if (status != GNOME_Evolution_Calendar_Success) {
		e_data_cal_view_notify_done (query, status);
                return;
	}

       	/* notify listeners of all objects */
	if (objects) {
		e_data_cal_view_notify_objects_added (query, (const GList *) objects);

		/* free memory */
		g_list_foreach (objects, (GFunc) g_free, NULL);
		g_list_free (objects);
	}

	e_data_cal_view_notify_done (query, GNOME_Evolution_Calendar_Success);
}

/* Set_mode handler for the file backend */
/* To support online/offline mode not implemented yet! */
static void
e_cal_backend_ogo_set_mode (ECalBackend *backend, CalMode mode)
{
	ECalBackendOGo *ogo;
	ECalBackendOGoPrivate *priv;

	ogo = E_CAL_BACKEND_OGO (backend);
	priv = ogo->priv;

	priv->mode = mode;	
	/* g_warning ("%s: implement me!", __PRETTY_FUNCTION__); */

}

/* is_remote handler for the file backend */
static CalMode
e_cal_backend_ogo_get_mode (ECalBackend *backend)
{
	ECalBackendOGo *ogo;
	ECalBackendOGoPrivate *priv;

	ogo = E_CAL_BACKEND_OGO (backend);
	priv = ogo->priv;

	return priv->mode;
}

/* is_loaded handler for the file backend */
static gboolean
e_cal_backend_ogo_is_loaded (ECalBackend *backend)
{
	ECalBackendOGo *ogo;
	ECalBackendOGoPrivate *priv;

	ogo = E_CAL_BACKEND_OGO (backend);
	priv = ogo->priv;

	return priv->cache ? TRUE : FALSE;
}

static void
e_cal_backend_ogo_dispose (GObject *object)
{
	ECalBackendOGo *ogo;
	ECalBackendOGoPrivate *priv;

	ogo = E_CAL_BACKEND_OGO (object);
	priv = ogo->priv;

	if (priv->mutex) {
		g_mutex_free (priv->mutex);
		priv->mutex = NULL;
	}
	
	if (priv->cache) {
		g_object_unref (priv->cache);
		priv->cache = NULL;
	}

	if (priv->timeout_id) {
		g_source_remove (priv->timeout_id);
		priv->timeout_id = 0;
	}

	if (priv->username) {
		g_free (priv->username);
		priv->username = NULL;
	}
	
	if (priv->password) {
		g_free (priv->password);
		priv->password = NULL;
	}

	if (G_OBJECT_CLASS (parent_class)->dispose)
		(* G_OBJECT_CLASS (parent_class)->dispose) (object);
}

/* Finalize handler for the file backend */
static void
e_cal_backend_ogo_finalize (GObject *object)
{
	ECalBackendOGo *ogo;
	ECalBackendOGoPrivate *priv;

	g_return_if_fail (object != NULL);
	g_return_if_fail (E_IS_CAL_BACKEND_OGO (object));

	ogo = E_CAL_BACKEND_OGO (object);
	priv = ogo->priv;

	if (G_OBJECT_CLASS (parent_class)->finalize)
		(* G_OBJECT_CLASS (parent_class)->finalize) (object);
}

static void
e_cal_backend_ogo_class_init (ECalBackendOGoClass *class)
{
	GObjectClass *object_class;
	ECalBackendClass *backend_class;
	ECalBackendSyncClass *sync_class;

	object_class = (GObjectClass *) class;
	backend_class = (ECalBackendClass *) class;
	sync_class = (ECalBackendSyncClass *) class;

	parent_class = g_type_class_peek_parent (class);

	object_class->dispose = e_cal_backend_ogo_dispose;
	object_class->finalize = e_cal_backend_ogo_finalize;

	sync_class->is_read_only_sync = e_cal_backend_ogo_is_read_only;
	sync_class->get_cal_address_sync = e_cal_backend_ogo_get_cal_address;
 	sync_class->get_alarm_email_address_sync = e_cal_backend_ogo_get_alarm_email_address;
 	sync_class->get_ldap_attribute_sync = e_cal_backend_ogo_get_ldap_attribute;
 	sync_class->get_static_capabilities_sync = e_cal_backend_ogo_get_static_capabilities;
	sync_class->open_sync = e_cal_backend_ogo_open;
	sync_class->remove_sync = e_cal_backend_ogo_remove;
	sync_class->create_object_sync = e_cal_backend_ogo_create_object;
	sync_class->modify_object_sync = e_cal_backend_ogo_modify_object;
	sync_class->remove_object_sync = e_cal_backend_ogo_remove_object;
	sync_class->discard_alarm_sync = e_cal_backend_ogo_discard_alarm;
	sync_class->receive_objects_sync = e_cal_backend_ogo_receive_objects;
	sync_class->send_objects_sync = e_cal_backend_ogo_send_objects;
 	sync_class->get_default_object_sync = e_cal_backend_ogo_get_default_object;
	sync_class->get_object_sync = e_cal_backend_ogo_get_object;
	sync_class->get_object_list_sync = e_cal_backend_ogo_get_object_list;
	sync_class->get_timezone_sync = e_cal_backend_ogo_get_timezone;
	sync_class->add_timezone_sync = e_cal_backend_ogo_add_timezone;
	sync_class->set_default_timezone_sync = e_cal_backend_ogo_set_default_timezone;
	sync_class->get_freebusy_sync = e_cal_backend_ogo_get_free_busy;
	sync_class->get_changes_sync = e_cal_backend_ogo_get_changes;

	backend_class->is_loaded = e_cal_backend_ogo_is_loaded;
	backend_class->start_query = e_cal_backend_ogo_start_query;
	backend_class->get_mode = e_cal_backend_ogo_get_mode;
	backend_class->set_mode = e_cal_backend_ogo_set_mode;
	backend_class->internal_get_default_timezone = e_cal_backend_ogo_internal_get_default_timezone;
	backend_class->internal_get_timezone = e_cal_backend_ogo_internal_get_timezone;

	g_type_class_add_private (object_class, sizeof (ECalBackendOGoPrivate));
	
}

static void
e_cal_backend_ogo_init (ECalBackendOGo      *ogo, 
			ECalBackendOGoClass *class)
{
	ECalBackendOGoPrivate *priv;
	
	ogo->priv = G_TYPE_INSTANCE_GET_PRIVATE (ogo, E_TYPE_CAL_BACKEND_OGO,
						 ECalBackendOGoPrivate);

	/* create the mutex for thread safety */
	ogo->priv->mutex = g_mutex_new ();
}

/**
 * e_cal_backend_ogo_get_type:
 * @void: 
 * 
 * Registers the #ECalBackendOGo class if necessary, and returns the type ID
 * associated to it.
 * 
 * Return value: The type ID of the #ECalBackendOGo class.
 **/
GType
e_cal_backend_ogo_get_type (void)
{
	static GType type = 0;

	if (!type) {
		static GTypeInfo info = {
                        sizeof (ECalBackendOGoClass),
                        (GBaseInitFunc) NULL,
                        (GBaseFinalizeFunc) NULL,
                        (GClassInitFunc) e_cal_backend_ogo_class_init,
                        NULL, NULL,
                        sizeof (ECalBackendOGo),
                        0,
                        (GInstanceInitFunc) e_cal_backend_ogo_init
                };
		type = g_type_register_static (E_TYPE_CAL_BACKEND_SYNC,
					       "ECalBackendOGo", 
					       &info, 0);
	}

	return type;
}
