/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * OpenGroupware.org connection library
 *
 *
 * Authors : 
 *  JP Rosevear <jpr@ximian.com>
 *  Rodrigo Moya <rodrigo@ximian.com>
 *  Anders Carlsson <andersca@imendio.com>
 *
 * Copyright 2003, Novell, Inc.
 * Copyright 2004, Imendio AB
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */


#include <libsoup/soup-session-sync.h>
#include <libsoup/soup-uri.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <libedataserver/e-url.h>
#include "e2k-encoding-utils.h"
#include "e2k-xml-utils.h"
#include "e2k-uri.h"
#include "e2k-utils.h"
#include "ogo-connection.h"

static GObjectClass *parent_class       = NULL;
static GHashTable   *loaded_connections = NULL;

struct _OGoConnectionPrivate {
	SoupSession *soup_session;

	char *uri;
	char *username;
	char *password;
};

static void
ogo_connection_authenticate (SoupSession *session, 
			     SoupMessage *msg,
			     const char  *auth_type, 
			     const char  *auth_realm,
			     char       **username, 
			     char       **password, 
			     gpointer     data)
{
	OGoConnection *conn = data;

	*username = g_strdup (conn->priv->username);
	*password = g_strdup (conn->priv->password);
}

static void
ogo_connection_reauthenticate (SoupSession *session, 
			       SoupMessage *msg, 
			       const char  *auth_type, 
			       const char  *auth_realm,
			       char       **username, 
			       char       **password, 
			       gpointer     data)
{
	OGoConnection *conn = data;

	*username = g_strdup (conn->priv->username);
	*password = g_strdup (conn->priv->password);
	
	g_print ("Warning: reauthenticate!\n");

}

static void
ogo_connection_finalize (GObject *object)
{
	OGoConnection *conn = OGO_CONNECTION (object);

	g_free (conn->priv->uri);
	g_free (conn->priv->username);
	g_free (conn->priv->password);

	g_object_unref (conn->priv->soup_session);
	
	(* parent_class->finalize) (object);
}

 
static void
ogo_connection_class_init (OGoConnectionClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	parent_class = g_type_class_peek_parent (klass);

	object_class->finalize = ogo_connection_finalize;
	
	g_type_class_add_private (object_class, sizeof (OGoConnectionPrivate));

}

static void
ogo_connection_init (OGoConnection *conn)
{
	OGoConnectionPrivate *priv;

	priv = conn->priv = G_TYPE_INSTANCE_GET_PRIVATE 
	  (conn, OGO_TYPE_CONNECTION, OGoConnectionPrivate);

	/* create the SoupSession for this connection */
	priv->soup_session = soup_session_sync_new ();
	g_signal_connect (priv->soup_session, "authenticate",
			  G_CALLBACK (ogo_connection_authenticate), conn);
	g_signal_connect (priv->soup_session, "reauthenticate",
			  G_CALLBACK (ogo_connection_reauthenticate), conn);
}

GType
ogo_connection_get_type (void)
{
	static GType type = 0;

	if (!type) {
		static GTypeInfo info = {
                        sizeof (OGoConnectionClass),
                        (GBaseInitFunc) NULL,
                        (GBaseFinalizeFunc) NULL,
                        (GClassInitFunc) ogo_connection_class_init,
                        NULL, NULL,
                        sizeof (OGoConnection),
                        0,
                        (GInstanceInitFunc) ogo_connection_init
                };
		type = g_type_register_static (G_TYPE_OBJECT, "OGoConnection", &info, 0);
	}

	return type;
}

OGoConnection *
ogo_connection_new (const char *uri, 
		    const char *username, 
		    const char *password)
{
	OGoConnection       *conn;
	OGoConnectionStatus  status;
	char                *hash_key;
	static GStaticMutex  connecting = G_STATIC_MUTEX_INIT;	
	
	g_static_mutex_lock (&connecting);

	/* search the connection in our hash table */
	if (loaded_connections != NULL) {
		hash_key = g_strdup_printf ("%s:%s@%s",
					    username ? username : "",
					    password ? password : "",
					    uri);
		conn = g_hash_table_lookup (loaded_connections, hash_key);
		g_free (hash_key);

		if (OGO_IS_CONNECTION (conn)) {
			g_object_ref (conn);
			g_static_mutex_unlock (&connecting);
			return conn;
		}
	}

	
	/* not found, so create a new connection */
	conn = g_object_new (OGO_TYPE_CONNECTION, NULL);

	conn->priv->uri =      g_strdup (uri);
	conn->priv->username = g_strdup (username);
	conn->priv->password = g_strdup (password);

	/* add the connection to the loaded_connections hash table */
	hash_key = g_strdup_printf ("%s:%s@%s",
				    conn->priv->username ? conn->priv->username : "",
				    conn->priv->password ? conn->priv->password : "",
				    conn->priv->uri);
	if (loaded_connections == NULL)
		loaded_connections = g_hash_table_new (g_str_hash, g_str_equal);
	g_hash_table_insert (loaded_connections, hash_key, conn);

	g_static_mutex_unlock (&connecting);

	return conn;
}

static OGoConnectionStatus
get_status_from_http_response (guint response)
{
	switch (response) {
	case SOUP_STATUS_OK:
	case SOUP_STATUS_MULTI_STATUS:
	case SOUP_STATUS_NO_CONTENT:
	case SOUP_STATUS_CREATED:
		return OGO_CONNECTION_STATUS_OK;
	case SOUP_STATUS_NOT_FOUND:
		return OGO_CONNECTION_STATUS_NOT_FOUND;
	case SOUP_STATUS_UNAUTHORIZED:
		return OGO_CONNECTION_STATUS_ACCESS_DENIED;
	default:
		g_warning ("Unknown response code %d", response);
		return OGO_CONNECTION_STATUS_UNKNOWN;
	}
}

static void
print_header (gpointer name, 
	      gpointer value, 
	      gpointer data)
{
	g_print ("%s: %s\n", (char *) name, (char *) value);
}

static void
debug_handler (SoupMessage *msg, 
	       gpointer     user_data)
{
	g_print ("%d %s\nSOAP-Debug: %p @ %lu\n",
                msg->status_code, msg->reason_phrase,
                msg, time (0));

	/* print headers */
	soup_message_foreach_header (msg->response_headers, print_header, NULL);

	/* print response */
 	if (msg->response.length) {
		fputc ('\n', stdout);
		fwrite (msg->response.body, 1, msg->response.length, stdout);
		fputc ('\n', stdout);
	}
}

static void
setup_debug (SoupMessage *msg)
{
	const SoupUri *suri;

	suri = soup_message_get_uri (SOUP_MESSAGE (msg));
	g_print ("%s %s%s%s HTTP/1.1\nSOAP-Debug: %p @ %lu\n",
		 SOUP_MESSAGE (msg)->method, suri->path,
		 suri->query ? "?" : "",
		 suri->query ? suri->query : "",
		 msg, (unsigned long) time (0));

	/* print message headers */
	print_header ("Host", suri->host, NULL);
	soup_message_foreach_header (SOUP_MESSAGE (msg)->request_headers, print_header, NULL);

 	if (msg->request.length) {
		fputc ('\n', stdout);
		fwrite (msg->request.body, 1, msg->request.length, stdout);
		fputc ('\n', stdout);
	}
	
	soup_message_add_handler (SOUP_MESSAGE (msg), SOUP_HANDLER_POST_BODY, 
				  debug_handler, NULL);
}
static SoupMessage *
ogo_message_new (const gchar *method,
		 const gchar *uri)
{
	SoupMessage *message;

	message = soup_message_new (method, uri);

	return message;
}

OGoConnectionStatus
ogo_connection_delete_item (OGoConnection *conn,
			    const gchar   *uri,
			    const gchar   *etag)
{
	EUri *parsed_uri;
	SoupMessage *message;
	guint http_status;
	gchar *real_uri;
	
	/* FIXME: We assume here that the item hrefs point to the same server which might not be true */
	parsed_uri = e_uri_new (uri);
	real_uri = g_strdup_printf ("%s%s",
				    conn->priv->uri, parsed_uri->path);
	e_uri_free (parsed_uri);

	message = soup_message_new (SOUP_METHOD_DELETE, real_uri);
	g_free (real_uri);
	soup_message_add_header (message->request_headers, "If-match", etag);

	http_status = soup_session_send_message (conn->priv->soup_session, message);
	g_object_unref (message);
	
	return get_status_from_http_response (http_status);
}

OGoConnectionStatus
ogo_connection_modify_item (OGoConnection *conn,
			    const gchar   *uri,
			    const gchar   *new_contents,
			    const gchar   *etag,
			    const gchar   *mime_type)
{
	EUri         *parsed_uri;
	gchar        *real_uri;
	SoupMessage  *message;
	guint         http_status;
	
	/* FIXME: We assume here that the item hrefs point to the same server which might not be true */
	parsed_uri = e_uri_new (uri);
	real_uri = g_strdup_printf ("%s%s",
				    conn->priv->uri, parsed_uri->path);
	e_uri_free (parsed_uri);
	
	message = ogo_message_new (SOUP_METHOD_PUT, real_uri);
	g_free (real_uri);

	soup_message_add_header (message->request_headers, "If-match", etag);
	g_print ("new request: %s\n", new_contents);
	
	soup_message_set_request (message, mime_type, SOUP_BUFFER_STATIC,
	 			  (gchar *)new_contents, strlen (new_contents));

	setup_debug (message);
	
	http_status = soup_session_send_message (conn->priv->soup_session, message);
	g_object_unref (message);

	return get_status_from_http_response (http_status);
}


OGoConnectionStatus
ogo_connection_fetch_item_absolute (OGoConnection *conn,
				    const gchar   *absolute_path,
				    gchar        **item_contents)
{
	SoupMessage         *message;
	guint                http_status;
	OGoConnectionStatus  status;
	gchar               *real_uri;

	real_uri = g_strdup_printf ("%s%s",
				     conn->priv->uri, absolute_path);

	message = ogo_message_new (SOUP_METHOD_GET, real_uri);
	g_free (real_uri);
	
	http_status = soup_session_send_message (conn->priv->soup_session, message);

	if ((status = get_status_from_http_response (http_status)) == OGO_CONNECTION_STATUS_OK) {
		*item_contents = g_strndup (message->response.body, 
					    message->response.length);
		/* *item_contents = g_locale_to_utf8 (message->response.body, 
		   message->response.length, NULL, NULL, NULL); */
		/* g_message ("CACHE: %s\n", *item_contents); */
	} else {
	  *item_contents = NULL;
	}

	g_object_unref (message);

	return status;
}


static xmlNode *
find_first_child (xmlNode     *parent,
		  const gchar *name)
{
	xmlNode *node;
	
	for (node = parent->children; node != NULL; node = node->next) {
		if (strcmp (node->name, name) == 0)
			return node;
	}
	
	return NULL;
}
		  
static void
parse_list_response (xmlDocPtr     doc,
		     GList       **items)
{
	xmlNode *root, *response, *propstat, *prop, *value, *uri_node;
	gchar   *display_name, *uri; 
	GList   *tmp = NULL;
	
	root = xmlDocGetRootElement (doc);

	for (response = root->children; response; response = response->next) {
		OGoItem *item;

		if (response->type != XML_ELEMENT_NODE) {
			continue;
		}

		item = g_new0 (OGoItem, 1);
		
		uri_node = find_first_child (response, "href");
		if (uri_node->children->content)
			item->uri = g_strdup (uri_node->children->content);
		
		propstat = find_first_child (response, "propstat");
		prop = find_first_child (propstat, "prop");

		display_name = NULL;
		
		/* Now walk through all properties */
		for (value = prop->children; value != NULL; value = value->next) {
			if (strcmp (value->name, "getetag") == 0) {
				item->etag = g_strdup (value->children->content);
			}
		}

		tmp = g_list_prepend (tmp, item);
	}
	*items = tmp;	
}

OGoConnectionStatus
ogo_connection_list_items (OGoConnection *conn,
			   const gchar   *item_path,
			   GList        **items)
{
	SoupMessage         *message;
	guint                http_status;
	OGoConnectionStatus  status;
	gchar               *real_uri;
	const gchar         *req = "<?xml version=\"1.0\" encoding=\"utf-8\"?>"
		                   "<propfind xmlns=\"DAV:\"><prop><getetag/></prop></propfind>";
	
	real_uri = g_strdup_printf ("%s/zidestore/dav/%s/%s",
				     conn->priv->uri, conn->priv->username, item_path);
	message = ogo_message_new (SOUP_METHOD_PROPFIND, real_uri);
	soup_message_set_request (message, "text/xml", SOUP_BUFFER_STATIC,
				  (gchar *)req, strlen (req));
	http_status = soup_session_send_message (conn->priv->soup_session, message);

	if ((status = get_status_from_http_response (http_status)) == OGO_CONNECTION_STATUS_OK) {
		xmlDocPtr doc;

		doc = xmlReadMemory (message->response.body, message->response.length, 
				     NULL, NULL, XML_PARSE_NOWARNING);
		parse_list_response (doc, items);
		xmlFreeDoc (doc);
	} else {
		*items = NULL;
	}

	g_object_unref (message);
	
	g_free (real_uri);
	
	return status;
}

OGoConnectionStatus
ogo_connection_fetch_items (OGoConnection  *conn,
			    const gchar    *item_path,
			    GList        **items)
{
	OGoConnectionStatus  status;
	GList               *tmp_items, *list;

	status = ogo_connection_list_items (conn, item_path, &tmp_items);

	if (status != OGO_CONNECTION_STATUS_OK)
		return status;

	for (list = tmp_items; list != NULL; list = list->next) {
		OGoItem *item = list->data;
		EUri *uri;
      
		/* FIXME: We assume here that the item hrefs point to the same server which might not be true */
		uri = e_uri_new (item->uri);
		ogo_connection_fetch_item_absolute (conn, uri->path, &item->contents);
		e_uri_free (uri);
	}
	*items = tmp_items;
	return OGO_CONNECTION_STATUS_OK;
}

static void
ogo_item_free (OGoItem *item)
{
	g_free (item->uri);
	g_free (item->contents);
	g_free (item->uid);
	g_free (item);
}

void
ogo_item_list_free (GList *items)
{
	GList *tmp;
	
	for (tmp = items; tmp != NULL; tmp = tmp->next) {
		OGoItem *item = tmp->data;

		ogo_item_free (item);
	}
	g_list_free (items);
}


static SoupMessage *
propfind_msg (const char  *base_uri,
	      const char **props, 
	      int          nprops,
	      const char **hrefs, 
	      int          nhrefs)
{
	SoupMessage *msg;
	GString     *propxml;
	GData       *set_namespaces;
	const char  *name;
	char         abbrev;
	int          i;

	propxml = g_string_new (E2K_XML_HEADER);
	g_string_append (propxml, "<D:propfind xmlns:D=\"DAV:\"");

	set_namespaces = NULL;
	for (i = 0; i < nprops; i++) {
		name = e2k_prop_namespace_name (props[i]);
		abbrev = e2k_prop_namespace_abbrev (props[i]);

		if (!g_datalist_get_data (&set_namespaces, name)) {
			g_datalist_set_data (&set_namespaces, name,
					     GINT_TO_POINTER (1));
			g_string_append_printf (propxml, " xmlns:%c=\"%s\"",
						abbrev, name);
		}
	}
	g_datalist_clear (&set_namespaces);
	g_string_append (propxml, ">\r\n");

	if (hrefs) {
		g_string_append (propxml, "<D:target>\r\n");
		for (i = 0; i < nhrefs; i++) {
			g_string_append_printf (propxml, "<D:href>%s</D:href>",
						hrefs[i]);
		}
		g_string_append (propxml, "\r\n</D:target>\r\n");
	}

	g_string_append (propxml, "<D:prop>\r\n");
	for (i = 0; i < nprops; i++) {
		abbrev = e2k_prop_namespace_abbrev (props[i]);
		name = e2k_prop_property_name (props[i]);
		g_string_append_printf (propxml, "<%c:%s/>", abbrev, name);
	}
	g_string_append (propxml, "\r\n</D:prop>\r\n</D:propfind>");

	msg = ogo_message_new (SOUP_METHOD_PROPFIND, base_uri);
	soup_message_set_request (msg, "text/xml", SOUP_BUFFER_SYSTEM_OWNED,
				  propxml->str, propxml->len);
	g_string_free (propxml, FALSE);
	soup_message_add_header (msg->request_headers, "Brief", "t");
	soup_message_add_header (msg->request_headers, "Depth", "0");

	return msg;
}

OGoConnectionStatus
ogo_connection_propfind (OGoConnection  *conn,
			 const gchar    *uri,
			 const gchar   **props,
			 gint            nprops,
			 E2kProperties **out_props,
			 gint           *nout_props)
{
	SoupMessage *message;
	gint http_status;
	EUri *parsed_uri;
	gint n_results;
	
	message = propfind_msg (uri, props, nprops, NULL, 0);
	soup_message_add_header (message->request_headers, "User-Agent",
				 "Evolution/2.0.1");
	g_print ("message: %p\n", message);
	setup_debug (message);

	http_status = soup_session_send_message (conn->priv->soup_session, message);
	g_object_unref (message);
	
	return OGO_CONNECTION_STATUS_OK;
}


static void
add_namespaces (const char *namespace, 
		char        abbrev, 
		gpointer    user_data)
{
	GString *propxml = user_data;

	g_string_append_printf (propxml, " xmlns:%c=\"%s\"", abbrev, namespace);
}

static void
write_prop (GString    *xml, 
	    const char *propertyname,
	    E2kPropType type, 
	    gpointer    value, 
	    gboolean    set)
{
	const char *namespace, *name, *typestr;
	char       *encoded, abbrev;
	gboolean    b64enc, need_type;
	GByteArray *data;
	GPtrArray  *array;
	int i;

	if (value == NULL)
		return;
	namespace = e2k_prop_namespace_name (propertyname);
	abbrev = e2k_prop_namespace_abbrev (propertyname);
	name = e2k_prop_property_name (propertyname);

	need_type = (strstr (namespace, "/mapi/id/") != NULL);

	g_string_append_printf (xml, "<%c:%s", abbrev, name);
	if (!set) {
		g_string_append (xml, "/>");
		return;
	} else if (!need_type)
		g_string_append_c (xml, '>');

	switch (type) {
	case E2K_PROP_TYPE_BINARY:
		if (need_type)
			g_string_append (xml, " T:dt=\"bin.base64\">");
		data = value;
		encoded = e2k_base64_encode (data->data, data->len);
		g_string_append (xml, encoded);
		g_free (encoded);
		break;

	case E2K_PROP_TYPE_STRING_ARRAY:
		typestr = " T:dt=\"mv.string\">";
		b64enc = FALSE;
		goto array_common;

	case E2K_PROP_TYPE_INT_ARRAY:
		typestr = " T:dt=\"mv.int\">";
		b64enc = FALSE;
		goto array_common;

	case E2K_PROP_TYPE_BINARY_ARRAY:
		typestr = " T:dt=\"mv.bin.base64\">";
		b64enc = TRUE;

	array_common:
		if (need_type)
			g_string_append (xml, typestr);
		array = value;
		for (i = 0; i < array->len; i++) {
			g_string_append (xml, "<X:v>");

			if (b64enc) {
				data = array->pdata[i];
				encoded = e2k_base64_encode (data->data,
							     data->len);
				g_string_append (xml, encoded);
				g_free (encoded);
			} else
				e2k_g_string_append_xml_escaped (xml, array->pdata[i]);

			g_string_append (xml, "</X:v>");
		}
		break;

	case E2K_PROP_TYPE_XML:
		g_assert_not_reached ();
		break;

	case E2K_PROP_TYPE_STRING:
	default:
		if (need_type) {
			switch (type) {
			case E2K_PROP_TYPE_INT:
				typestr = " T:dt=\"int\">";
				break;
			case E2K_PROP_TYPE_BOOL:
				typestr = " T:dt=\"boolean\">";
				break;
			case E2K_PROP_TYPE_FLOAT:
				typestr = " T:dt=\"float\">";
				break;
			case E2K_PROP_TYPE_DATE:
				typestr = " T:dt=\"dateTime.tz\">";
				break;
			default:
				typestr = ">";
				break;
			}
			g_string_append (xml, typestr);
		}
		e2k_g_string_append_xml_escaped (xml, value);
		break;

	}

	g_string_append_printf (xml, "</%c:%s>", abbrev, name);
}

static void
add_set_props (const char *propertyname, 
	       E2kPropType type,
	       gpointer    value, 
	       gpointer    user_data)
{
	GString **props = user_data;

	if (!*props)
		*props = g_string_new (NULL);

	write_prop (*props, propertyname, type, value, TRUE);
}

static void
add_remove_props (const char *propertyname, 
		  E2kPropType type,
		  gpointer    value, 
		  gpointer    user_data)
{
	GString **props = user_data;

	if (!*props)
		*props = g_string_new (NULL);

	write_prop (*props, propertyname, type, value, FALSE);
}

static SoupMessage *
proppatch_msg (const char    *base_uri, 
	       E2kProperties *props,
	       gboolean       create)
{
	SoupMessage *msg;
	GString     *propxml, *subxml;
	int          i;

	propxml = g_string_new (E2K_XML_HEADER);
	g_string_append (propxml, "<D:propertyupdate xmlns:D=\"DAV:\"");

	/* Iterate over the properties, noting each namespace once,
	 * then add them all to the header.
	 */
	e2k_properties_foreach_namespace (props, add_namespaces, propxml);
	g_string_append (propxml, ">\r\n");

	/* Add <set> properties. */
	subxml = NULL;
	e2k_properties_foreach (props, add_set_props, &subxml);
	if (subxml) {
		g_string_append (propxml, "<D:set><D:prop>\r\n");
		g_string_append (propxml, subxml->str);
		g_string_append (propxml, "\r\n</D:prop></D:set>");
		g_string_free (subxml, TRUE);
	}

	/* Add <remove> properties. */
	subxml = NULL;
	e2k_properties_foreach_removed (props, add_remove_props, &subxml);
	if (subxml) {
		g_string_append (propxml, "<D:remove><D:prop>\r\n");
		g_string_append (propxml, subxml->str);
		g_string_append (propxml, "\r\n</D:prop></D:remove>");
		g_string_free (subxml, TRUE);
	}

	/* Finish it up */
	g_string_append (propxml, "\r\n</D:propertyupdate>");

	/* And build the message. */
	msg = ogo_message_new (SOUP_METHOD_PROPPATCH, base_uri);
	soup_message_set_request (msg, "text/xml", SOUP_BUFFER_SYSTEM_OWNED,
				  propxml->str, propxml->len);
	g_print ("MESSAGE IS: %s\n", propxml->str);
	g_string_free (propxml, FALSE);
	soup_message_add_header (msg->request_headers, "Brief", "t");
	if (!create)
		soup_message_add_header (msg->request_headers, "If-Match", "*");

	return msg;
}

static void
extract_put_results (SoupMessage  *msg, 
		     char        **location, 
		     char        **repl_uid)
{
	const char *header;

	if (repl_uid) {
		header = soup_message_get_header (msg->response_headers,
						  "Repl-UID");
		*repl_uid = g_strdup (header);
	}
	if (location) {
		header = soup_message_get_header (msg->response_headers,
						  "Location");
		*location = g_strdup (header);
	}
}

OGoConnectionStatus
ogo_connection_proppatch (OGoConnection *conn,
			  const char    *uri, 
			  E2kProperties *props,
			  gboolean       create, 
			  char         **repl_uid)
{
	SoupMessage         *message;
	gint                 http_status;
	EUri                *parsed_uri;
	OGoConnectionStatus  status;
	
	message = proppatch_msg (uri, props, create);
	soup_message_add_header (message->request_headers, "User-Agent",
				 "Evolution/2.0.1");
	
	setup_debug (message);

	http_status = soup_session_send_message (conn->priv->soup_session, message);

	status = get_status_from_http_response (http_status);
	
	if (status == OGO_CONNECTION_STATUS_OK)
		extract_put_results (message, NULL, repl_uid);

	g_object_unref (message);

	g_print ("http status: %d\n", http_status);
	
	return status;
}

static void
update_unique_uri (OGoConnection             *conn, 
		   SoupMessage               *msg,
		   const char                *folder_uri, 
		   const char                *encoded_name, 
		   int                       *count,
		   OGoConnectionTestCallback  test_callback, 
		   gpointer                   user_data)
{
	SoupUri *suri;
	char *uri = NULL;

	do {
		g_free (uri);
		if (*count == 1) {
			uri = g_strdup_printf ("%s%s.EML", folder_uri,
					       "1Sliffsloff");
#if 0			
			uri = g_strdup_printf ("%s%s.EML", folder_uri,
					       encoded_name);
#endif
		} else {
			uri = g_strdup_printf ("%s%s-%d.EML", folder_uri,
					       encoded_name, *count);
		}
		(*count)++;
	} while (test_callback && !test_callback (conn, uri, user_data));

	suri = soup_uri_new (uri);
	soup_message_set_uri (msg, suri);
	soup_uri_free (suri);
	g_free (uri);
}

OGoConnectionStatus
ogo_connection_proppatch_new (OGoConnection             *conn,
			      const char                *folder_uri,
			      const char                *object_name,
			      OGoConnectionTestCallback  test_callback,
			      gpointer                   user_data,
			      E2kProperties             *props,
			      char                     **location, 
			      char                     **repl_uid)
{
	SoupMessage         *msg;
	OGoConnectionStatus  status;
	char                *slash_uri, *encoded_name;
	int                  count;

	slash_uri = e2k_strdup_with_trailing_slash (folder_uri);
	encoded_name = e2k_uri_encode (object_name, TRUE, NULL);

	/* folder_uri is a dummy here */
	msg = proppatch_msg (folder_uri, props, TRUE);
	soup_message_add_header (msg->request_headers, "If-None-Match", "*");

	count = 1;
	do {
		update_unique_uri (conn, msg, slash_uri, encoded_name, &count,
				   test_callback, user_data);
		setup_debug (msg);
		status = soup_session_send_message (conn->priv->soup_session, msg);
	} while (status == 412);

	if (location)
		*location = soup_uri_to_string (soup_message_get_uri (msg), FALSE);
	extract_put_results (msg, NULL, repl_uid);

	g_object_unref (msg);
	g_free (slash_uri);
	g_free (encoded_name);
	return status;
  
}


OGoConnectionStatus
ogo_connection_create_item (OGoConnection *conn,
			    const gchar   *item_path,
			    const gchar   *name,
			    const gchar   *new_contents,
			    const gchar   *mime_type,
			    gchar        **new_uri)
{
	gchar       *real_uri;
	SoupMessage *message;
	guint        http_status;
	const gchar *location;
	
	real_uri = g_strdup_printf ("%s/zidestore/dav/%s/%s/%s",
				    conn->priv->uri, conn->priv->username, item_path,
				    name);
	
	message = ogo_message_new (SOUP_METHOD_PUT, real_uri);
	g_free (real_uri);

	soup_message_add_header (message->request_headers, "If-None-Match", "*");
	soup_message_set_request (message, mime_type, SOUP_BUFFER_STATIC,
				  (gchar *)new_contents, strlen (new_contents));

	setup_debug (message);

	http_status = soup_session_send_message (conn->priv->soup_session, message);
	
	location = soup_message_get_header (message->response_headers, "Location");
	
	if (new_uri)
		*new_uri = g_strdup (location);
	
	g_object_unref (message);
	
	return get_status_from_http_response (http_status);
}

typedef struct {
	gchar *username;
	gchar *password;
	int    counts;
} AuthContext;

static void
try_authenticate (SoupSession *session, 
		  SoupMessage *msg,
		  const char  *auth_type, 
		  const char  *auth_realm,
		  char       **username, 
		  char       **password, 
		  gpointer     data)
{
	AuthContext *context = data;

	*username = g_strdup (context->username);
	*password = g_strdup (context->password);
	
	g_print ("try authenticate\n");
}

OGoConnectionStatus
ogo_try_authenticate (const gchar *uri, 
		      const gchar *username,
		      const gchar *password)
{
	SoupSession *soup_session;
	AuthContext *context;
	gchar       *real_uri;
	gint         http_status;
	SoupMessage *message;
	const gchar *req = "<?xml version=\"1.0\" encoding=\"utf-8\"?>"
		           "<propfind xmlns=\"DAV:\"><prop><getetag/></prop></propfind>";
	
	soup_session = soup_session_sync_new ();
	context = g_new0 (AuthContext, 1);
	context->username = g_strdup (username);
	context->password = g_strdup (password);
	context->counts = 0;
	
	g_signal_connect (soup_session, "authenticate",
			  G_CALLBACK (try_authenticate), context);
	
	real_uri = g_strdup_printf ("%s/zidestore/dav/%s",
				     uri, context->username);
	message = ogo_message_new (SOUP_METHOD_PROPFIND, real_uri);
	g_free (real_uri);
	soup_message_set_request (message, "text/xml", SOUP_BUFFER_STATIC,
				  (gchar *)req, strlen (req));
	http_status = soup_session_send_message (soup_session, message);
	g_object_unref (message);
	
	g_free (context->username);
	g_free (context->password);
	g_free (context);

	return get_status_from_http_response (http_status);
}

gchar *
ogo_connection_get_item_uri (OGoConnection *conn,
			     const gchar   *item_path)
{
	gchar *real_uri;
	
	real_uri = g_strdup_printf ("%s/zidestore/dav/%s/%s",
				    conn->priv->uri, conn->priv->username, item_path);
	
	return real_uri;
}

     
